"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import random

APIKEY = 'lstzFDEOhfFNMLikKa0am9mgEKLBl49T'

ASSET_TYPE = ['etf', 'fund', 'stock']

EXCHANGE = {
            "NYSE": {
                    "name": "exchanges_equity_nyse",
                    "id": "E0EXG$XNYS"
                    },

            "NASDAQ" : {
                    "name": "exchanges_equity_nasdaq",
                    "id": "E0EXG$XNAS"
                        },

            "LSE" : {
                    "name": "exchanges_london_stock_exchange",
                    "id": "E0EXG$XLON"
                    },

            "AMSTERDAM" : {
                    "name": "exchanges_equity_nyse_euronext_amsterdam",
                    "id": "E0EXG$XAMS"
                        },
 
            "ATHENS" : {
                    "name": "exchanges_equity_athens_stock_exchange",
                    "id": "E0EXG$XATH"
                        },

            "BOLSA_DE_VALORES" : {
                    "name": "exchanges_equity_bolsa_de_valores",
                    "id": "E0EXG$XMEX"
                        },

            "BOMBAY" : {
                    "name": "exchanges_equity_bombay_stock_exchange",
                    "id": "E0EXG$XBOM"
                        },

            "BORSA_ITALIANA" : {
                    "name": "exchanges_equity_borsa_italiana",
                    "id": "E0EXG$XMIL"
                        },

            "BRUSSELS" : {
                    "name": "exchanges_equity_nyse_brussels",
                    "id": "E0EXG$XBRU"
                        },

            "COPENHAGEN" : {
                    "name": "exchanges_equity_omx_exchange_copenhagen",
                    "id": "E0EXG$XCSE"
                            },

            "HELSINKI" : {
                    "name": "exchanges_equity_omx_exchange_helsinki",
                    "id": "E0EXG$XHEL"
                        },

                "HONG-KONG" : {
                    "name": "exchanges_HONG_KONG",
                    "id": "E0EXG$XHKG"
                        },
            "ICELAND" : {
                    "name": "exchanges_equity_omx_exchange_iceland",
                    "id": "E0EXG$XICE"
                        },

            "INDIA" : {
                    "name": "exchanges_equity_india_stock_exchange",
                    "id": "E0EXG$XNSE"
                        },

            "IPSX" : {
                    "name": "exchanges_equity_exchange_IPSX",
                    "id": "E0EXG$IPSX"
                        },

            "IRELAND" : {
                    "name": "exchanges_equity_irish_stock_exchange",
                    "id": "E0EXG$XDUB"
                        },

            "ISTANBUL" : {
                    "name": "exchanges_equity_istanbul_stock_exchange",
                    "id": "E0EXG$XIST"
                        },

            "LISBON" : {
                    "name": "exchanges_equity_nyse_euronext_lisbon",
                    "id": "E0EXG$XLIS"
                        },

            "LUXEMBOURG" : {
                    "name": "exchanges_equity_luxembourg_stock_exchange",
                    "id": "E0EXG$XLUX"
                        },

            "OSLO_BORS" : {
                    "name": "exchanges_equity_oslo_bors",
                    "id": "E0EXG$XOSL"
                        },

            "PARIS" : {
                    "name": "exchanges_equity_nyse_paris",
                    "id": "E0EXG$XPAR"
                        },

            "RIGA" : {
                    "name": "exchanges_equity_omx_exchange_riga",
                    "id": "E0EXG$XRIS"
                        },

            "SHANGAI" : {
                    "name": "exchanges_equity_shanghai_stock_exchange",
                    "id": "E0EXG$XSHG"
                        },

            "SHENZHEN" : {
                    "name": "exchanges_equity_shenzhen_stock_exchange",
                    "id": "E0EXG$XSHE"
                        },

            "SINGAPORE" : {
                    "name": "exchanges_equity_singapore",
                    "id": "E0EXG$XSES"
                        },
        
            "STOCKHOLM" : {
                    "name": "exchanges_equity_nasdaq_omx_stockholm",
                    "id": "E0EXG$XSTO"
                        },

            "SWISS" : {
                    "name": "exchanges_equity_six__swiss_exchange",
                    "id": "E0EXG$XSWX"
                    },

            "TAIWAN" : {
                    "name": "exchanges_equity_taiwan_stock_exchange",
                    "id": "E0EXG$XTAI"
                        },

            "TALLIN" : {
                    "name": "exchanges_equity_tallin_stock_exchange",
                    "id": "E0EXG$XTAL"
                        },

            "THAILAND" : {
                    "name": "exchanges_equity_thailand_stock_exchange",
                    "id": "E0EXG$XBKK"
                        },

            "TOKYO" : {
                    "name": "exchanges_equity_tokyo_stock_exchange",
                    "id": "E0EXG$XTKS"
                        },

            "VILNIUS" : {
                    "name": "exchanges_equity_omx_exchange_vilnius",
                    "id": "E0EXG$XLIT"
                        },
                                                                                
            "WARSAW" :{
                    "name": "exchanges_equity_warsaw_stock_exchange",
                    "id": "E0EXG$XWAR"
                        },

            "WIENER_BOERSE" : {
                    "name": "exchanges_equity_wiener_boerse",
                    "id": "E0EXG$XWBO"
                        },

        "WORLDWIDE_EQUITY" : {
                    "name": "exchanges_worldwide_equity",
                    "id": "E0WWE$$ALL"
                        },
            }


FIELDS = [
    'AdministratorCompanyId',
    'AlphaM36',
    'AnalystRatingScale',
    'AverageCreditQualityCode',
    'AverageMarketCapital',
    'BetaM36',
    'BondStyleBox',
    'brandingCompanyId',
    'categoryId',
    'CategoryName',
    'ClosePrice',
    'currency',
    'DebtEquityRatio',
    'distribution',
    'DividendYield',
    'EBTMarginYear1',
    'EffectiveDuration',
    'EPSGrowth3YYear1',
    'equityStyle',
    'EquityStyleBox',
    'exchangeCode',
    'ExchangeId',
    'ExpertiseAdvanced',
    'ExpertiseBasic',
    'ExpertiseInformed',
    'FeeLevel',
    'fundShareClassId',
    'fundSize',
    'fundStyle',
    'FundTNAV',
    'GBRReturnD1',
    'GBRReturnM0',
    'GBRReturnM1',
    'GBRReturnM12',
    'GBRReturnM120',
    'GBRReturnM3',
    'GBRReturnM36',
    'GBRReturnM6',
    'GBRReturnM60',
    'GBRReturnW1',
    'geoRegion',
    'globalAssetClassId',
    'globalCategoryId',
    'iMASectorId',
    'IndustryName',
    'InitialPurchase',
    'instrumentName',
    'investment',
    'investmentExpertise',
    'investmentObjective',
    'investmentType',
    'investorType',
    'InvestorTypeEligibleCounterparty',
    'InvestorTypeProfessional',
    'InvestorTypeRetail',
    'LargestSector',
    'LegalName',
    'managementStyle',
    'ManagerTenure',
    'MarketCap',
    'MarketCountryName',
    'MaxDeferredLoad',
    'MaxFrontEndLoad',
    'MaximumExitCostAcquired',
    'MorningstarRiskM255',
    'Name',
    'NetMargin',
    'ongoingCharge',
    'OngoingCostActual',
    'PEGRatio',
    'PERatio',
    'PerformanceFeeActual',
    'PriceCurrency',
    'QuantitativeRating',
    'R2M36',
    'ReturnD1',
    'ReturnM0',
    'ReturnM1',
    'ReturnM12',
    'ReturnM120',
    'ReturnM3',
    'ReturnM36',
    'ReturnM6',
    'ReturnM60',
    'ReturnProfileGrowth',
    'ReturnProfileHedging',
    'ReturnProfileIncome',
    'ReturnProfileOther',
    'ReturnProfilePreservation',
    'ReturnW1',
    'RevenueGrowth3Y',
    'riskSrri',
    'ROATTM',
    'ROETTM',
    'ROEYear1',
    'ROICYear1',
    'SecId',
    'SectorName',
    'shareClassType',
    'SharpeM36',
    'StandardDeviationM36',
    'starRating',
    'StarRatingM255',
    'SustainabilityRank',
    'sustainabilityRating',
    'TenforeId',
    'Ticker',
    'totalReturn',
    'totalReturnTimeFrame',
    'TrackRecordExtension',
    'TransactionFeeActual',
    'umbrellaCompanyId',
    'Universe',
    'Yield_M12',
    'yieldPercent',

]


FILTER_FUND = [  
    'AdministratorCompanyId',
    'AnalystRatingScale',
    'BondStyleBox',
    'BrandingCompanyId',
    'CategoryId',
    'CollectedSRRI',
    'distribution',
    'EquityStyleBox',
    'ExpertiseInformed',
    'FeeLevel',
    'FundTNAV',
    'GBRReturnM0',
    'GBRReturnM12',
    'GBRReturnM120',
    'GBRReturnM36',
    'GBRReturnM60',
    'GlobalAssetClassId',
    'GlobalCategoryId',
    'IMASectorID',
    'IndexFund',
    'InvestorTypeProfessional',
    'LargestRegion',
    'LargestSector',
    'OngoingCharge',
    'QuantitativeRating',
    'ReturnProfilePreservation',
    'ShareClassTypeId',
    'starRating',
    'SustainabilityRank',
    'UmbrellaCompanyId',
    'Yield_M12',

        ]

FILTER_STOCK = [  
    'debtEquityRatio',
    'DividendYield',
    'epsGrowth3YYear1',
    'EquityStyleBox',
    'GBRReturnM0',
    'GBRReturnM12',
    'GBRReturnM36',
    'GBRReturnM60',
    'GBRReturnM120',
    'IndustryId',
    'MarketCap',
    'netMargin',
    'PBRatio',
    'PEGRatio',
    'PERatio',
    'PSRatio',
    'revenueGrowth3Y',
    'roattm',
    'roettm',
    'SectorId',
        ]



SITE = {
        
        'af' : {'iso3' : 'AFG', 'site' : ''},
        'ax' : {'iso3' : 'ALA', 'site' : ''},
        'al' : {'iso3' : 'ALB', 'site' : ''},
        'dz' : {'iso3' : 'DZA', 'site' : ''},
        'as' : {'iso3' : 'ASM', 'site' : ''},
        'ad' : {'iso3' : 'AND', 'site' : ''},
        'ao' : {'iso3' : 'AGO', 'site' : ''},
        'ai' : {'iso3' : 'AIA', 'site' : ''},
        'aq' : {'iso3' : 'ATA', 'site' : ''},
        'ag' : {'iso3' : 'ATG', 'site' : ''},
        'ar' : {'iso3' : 'ARG', 'site' : ''},
        'am' : {'iso3' : 'ARM', 'site' : ''},
        'aw' : {'iso3' : 'ABW', 'site' : ''},
        'au' : {'iso3' : 'AUS', 'site' : ''},
        'at' : {'iso3' : 'AUT', 'site' : ''},
        'az' : {'iso3' : 'AZE', 'site' : ''},
        'bs' : {'iso3' : 'BHS', 'site' : ''},
        'bh' : {'iso3' : 'BHR', 'site' : ''},
        'bd' : {'iso3' : 'BGD', 'site' : ''},
        'bb' : {'iso3' : 'BRB', 'site' : ''},
        'by' : {'iso3' : 'BLR', 'site' : ''},
        'be' : {'iso3' : 'BEL', 'site' : 'https://www.morningstar.be/be/'},
        'bz' : {'iso3' : 'BLZ', 'site' : ''},
        'bj' : {'iso3' : 'BEN', 'site' : ''},
        'bm' : {'iso3' : 'BMU', 'site' : ''},
        'bt' : {'iso3' : 'BTN', 'site' : ''},
        'bo' : {'iso3' : 'BOL', 'site' : ''},
        'bq' : {'iso3' : 'BES', 'site' : ''},
        'ba' : {'iso3' : 'BIH', 'site' : ''},
        'bw' : {'iso3' : 'BWA', 'site' : ''},
        'bv' : {'iso3' : 'BVT', 'site' : ''},
        'br' : {'iso3' : 'BRA', 'site' : 'https://www.morningstarbr.com/br/'},
        'io' : {'iso3' : 'IOT', 'site' : ''},
        'bn' : {'iso3' : 'BRN', 'site' : ''},
        'bg' : {'iso3' : 'BGR', 'site' : ''},
        'bf' : {'iso3' : 'BFA', 'site' : ''},
        'bi' : {'iso3' : 'BDI', 'site' : ''},
        'cv' : {'iso3' : 'CPV', 'site' : ''},
        'kh' : {'iso3' : 'KHM', 'site' : ''},
        'cm' : {'iso3' : 'CMR', 'site' : ''},
        'ca' : {'iso3' : 'CAN', 'site' : 'https://www.morningstar.ca/ca/'},
        'ky' : {'iso3' : 'CYM', 'site' : ''},
        'cf' : {'iso3' : 'CAF', 'site' : ''},
        'td' : {'iso3' : 'TCD', 'site' : ''},
        'cl' : {'iso3' : 'CHL', 'site' : 'https://www.morningstar.cl/cl/'},
        'cn' : {'iso3' : 'CHN', 'site' : ''},
        'cx' : {'iso3' : 'CXR', 'site' : ''},
        'cc' : {'iso3' : 'CCK', 'site' : ''},
        'co' : {'iso3' : 'COL', 'site' : ''},
        'km' : {'iso3' : 'COM', 'site' : ''},
        'cd' : {'iso3' : 'COD', 'site' : ''},
        'cg' : {'iso3' : 'COG', 'site' : ''},
        'ck' : {'iso3' : 'COK', 'site' : ''},
        'cr' : {'iso3' : 'CRI', 'site' : ''},
        'ci' : {'iso3' : 'CIV', 'site' : ''},
        'hr' : {'iso3' : 'HRV', 'site' : ''},
        'cu' : {'iso3' : 'CUB', 'site' : ''},
        'cw' : {'iso3' : 'CUW', 'site' : ''},
        'cy' : {'iso3' : 'CYP', 'site' : ''},
        'cz' : {'iso3' : 'CZE', 'site' : ''},
        'dk' : {'iso3' : 'DNK', 'site' : 'https://www.morningstar.dk/dk/'},
        'dj' : {'iso3' : 'DJI', 'site' : ''},
        'dm' : {'iso3' : 'DMA', 'site' : ''},
        'do' : {'iso3' : 'DOM', 'site' : ''},
        'ec' : {'iso3' : 'ECU', 'site' : ''},
        'eg' : {'iso3' : 'EGY', 'site' : ''},
        'sv' : {'iso3' : 'SLV', 'site' : ''},
        'gq' : {'iso3' : 'GNQ', 'site' : ''},
        'er' : {'iso3' : 'ERI', 'site' : ''},
        'ee' : {'iso3' : 'EST', 'site' : ''},
        'sz' : {'iso3' : 'SWZ', 'site' : ''},
        'et' : {'iso3' : 'ETH', 'site' : ''},
        'fk' : {'iso3' : 'FLK', 'site' : ''},
        'fo' : {'iso3' : 'FRO', 'site' : ''},
        'fj' : {'iso3' : 'FJI', 'site' : ''},
        'fi' : {'iso3' : 'FIN', 'site' : 'https://www.morningstar.fi/fi/'},
        'fr' : {'iso3' : 'FRA', 'site' : 'https://www.morningstar.fr/fr/'},
        'gf' : {'iso3' : 'GUF', 'site' : ''},
        'pf' : {'iso3' : 'PYF', 'site' : ''},
        'tf' : {'iso3' : 'ATF', 'site' : ''},
        'ga' : {'iso3' : 'GAB', 'site' : ''},
        'gm' : {'iso3' : 'GMB', 'site' : ''},
        'ge' : {'iso3' : 'GEO', 'site' : ''},
        'de' : {'iso3' : 'DEU', 'site' : 'https://www.morningstar.de/de/'},
        'gh' : {'iso3' : 'GHA', 'site' : ''},
        'gi' : {'iso3' : 'GIB', 'site' : ''},
        'gr' : {'iso3' : 'GRC', 'site' : ''},
        'gl' : {'iso3' : 'GRL', 'site' : ''},
        'gd' : {'iso3' : 'GRD', 'site' : ''},
        'gp' : {'iso3' : 'GLP', 'site' : ''},
        'gu' : {'iso3' : 'GUM', 'site' : ''},
        'gt' : {'iso3' : 'GTM', 'site' : ''},
        'gg' : {'iso3' : 'GGY', 'site' : ''},
        'gn' : {'iso3' : 'GIN', 'site' : ''},
        'gw' : {'iso3' : 'GNB', 'site' : ''},
        'gy' : {'iso3' : 'GUY', 'site' : ''},
        'ht' : {'iso3' : 'HTI', 'site' : ''},
        'hm' : {'iso3' : 'HMD', 'site' : ''},
        'va' : {'iso3' : 'VAT', 'site' : ''},
        'hn' : {'iso3' : 'HND', 'site' : ''},
        'hk' : {'iso3' : 'HKG', 'site' : 'https://www.morningstar.hk/hk/'},
        'hu' : {'iso3' : 'HUN', 'site' : ''},
        'is' : {'iso3' : 'ISL', 'site' : ''},
        'in' : {'iso3' : 'IND', 'site' : ''},
        'id' : {'iso3' : 'IDN', 'site' : ''},
        'ir' : {'iso3' : 'IRN', 'site' : ''},
        'iq' : {'iso3' : 'IRQ', 'site' : ''},
        'ie' : {'iso3' : 'IRL', 'site' : 'https://www.morningstarfunds.ie/ie/'},
        'im' : {'iso3' : 'IMN', 'site' : ''},
        'il' : {'iso3' : 'ISR', 'site' : ''},
        'it' : {'iso3' : 'ITA', 'site' : 'https://www.morningstar.it/it/'},
        'jm' : {'iso3' : 'JAM', 'site' : ''},
        'jp' : {'iso3' : 'JPN', 'site' : ''},
        'je' : {'iso3' : 'JEY', 'site' : ''},
        'jo' : {'iso3' : 'JOR', 'site' : ''},
        'kz' : {'iso3' : 'KAZ', 'site' : ''},
        'ke' : {'iso3' : 'KEN', 'site' : ''},
        'ki' : {'iso3' : 'KIR', 'site' : ''},
        'kp' : {'iso3' : 'PRK', 'site' : ''},
        'kr' : {'iso3' : 'KOR', 'site' : ''},
        'kw' : {'iso3' : 'KWT', 'site' : ''},
        'kg' : {'iso3' : 'KGZ', 'site' : ''},
        'la' : {'iso3' : 'LAO', 'site' : ''},
        'lv' : {'iso3' : 'LVA', 'site' : ''},
        'lb' : {'iso3' : 'LBN', 'site' : ''},
        'ls' : {'iso3' : 'LSO', 'site' : ''},
        'lr' : {'iso3' : 'LBR', 'site' : ''},
        'ly' : {'iso3' : 'LBY', 'site' : ''},
        'li' : {'iso3' : 'LIE', 'site' : ''},
        'lt' : {'iso3' : 'LTU', 'site' : ''},
        'lu' : {'iso3' : 'LUX', 'site' : ''},
        'mo' : {'iso3' : 'MAC', 'site' : ''},
        'mk' : {'iso3' : 'MKD', 'site' : ''},
        'mg' : {'iso3' : 'MDG', 'site' : ''},
        'mw' : {'iso3' : 'MWI', 'site' : ''},
        'my' : {'iso3' : 'MYS', 'site' : ''},
        'mv' : {'iso3' : 'MDV', 'site' : ''},
        'ml' : {'iso3' : 'MLI', 'site' : ''},
        'mt' : {'iso3' : 'MLT', 'site' : ''},
        'mh' : {'iso3' : 'MHL', 'site' : ''},
        'mq' : {'iso3' : 'MTQ', 'site' : ''},
        'mr' : {'iso3' : 'MRT', 'site' : ''},
        'mu' : {'iso3' : 'MUS', 'site' : ''},
        'yt' : {'iso3' : 'MYT', 'site' : ''},
        'mx' : {'iso3' : 'MEX', 'site' : 'https://www.morningstar.com.mx/mx/'},
        'fm' : {'iso3' : 'FSM', 'site' : ''},
        'md' : {'iso3' : 'MDA', 'site' : ''},
        'mc' : {'iso3' : 'MCO', 'site' : ''},
        'mn' : {'iso3' : 'MNG', 'site' : ''},
        'me' : {'iso3' : 'MNE', 'site' : ''},
        'ms' : {'iso3' : 'MSR', 'site' : ''},
        'ma' : {'iso3' : 'MAR', 'site' : ''},
        'mz' : {'iso3' : 'MOZ', 'site' : ''},
        'mm' : {'iso3' : 'MMR', 'site' : ''},
        'na' : {'iso3' : 'NAM', 'site' : ''},
        'nr' : {'iso3' : 'NRU', 'site' : ''},
        'np' : {'iso3' : 'NPL', 'site' : ''},
        'nl' : {'iso3' : 'NLD', 'site' : 'https://www.morningstar.nl/nl/'},
        'nc' : {'iso3' : 'NCL', 'site' : ''},
        'nz' : {'iso3' : 'NZL', 'site' : ''},
        'ni' : {'iso3' : 'NIC', 'site' : ''},
        'ne' : {'iso3' : 'NER', 'site' : ''},
        'ng' : {'iso3' : 'NGA', 'site' : ''},
        'nu' : {'iso3' : 'NIU', 'site' : ''},
        'nf' : {'iso3' : 'NFK', 'site' : ''},
        'mp' : {'iso3' : 'MNP', 'site' : ''},
        'no' : {'iso3' : 'NOR', 'site' : 'https://www.morningstar.no/no/'},
        'om' : {'iso3' : 'OMN', 'site' : ''},
        'pk' : {'iso3' : 'PAK', 'site' : ''},
        'pw' : {'iso3' : 'PLW', 'site' : ''},
        'ps' : {'iso3' : 'PSE', 'site' : ''},
        'pa' : {'iso3' : 'PAN', 'site' : ''},
        'pg' : {'iso3' : 'PNG', 'site' : ''},
        'py' : {'iso3' : 'PRY', 'site' : ''},
        'pe' : {'iso3' : 'PER', 'site' : ''},
        'ph' : {'iso3' : 'PHL', 'site' : ''},
        'pn' : {'iso3' : 'PCN', 'site' : ''},
        'pl' : {'iso3' : 'POL', 'site' : ''},
        'pt' : {'iso3' : 'PRT', 'site' : 'https://www.morningstar.pt/pt/'},
        'pr' : {'iso3' : 'PRI', 'site' : ''},
        'qa' : {'iso3' : 'QAT', 'site' : ''},
        're' : {'iso3' : 'REU', 'site' : ''},
        'ro' : {'iso3' : 'ROU', 'site' : ''},
        'ru' : {'iso3' : 'RUS', 'site' : ''},
        'rw' : {'iso3' : 'RWA', 'site' : ''},
        'bl' : {'iso3' : 'BLM', 'site' : ''},
        'sh' : {'iso3' : 'SHN', 'site' : ''},
        'kn' : {'iso3' : 'KNA', 'site' : ''},
        'lc' : {'iso3' : 'LCA', 'site' : ''},
        'mf' : {'iso3' : 'MAF', 'site' : ''},
        'pm' : {'iso3' : 'SPM', 'site' : ''},
        'vc' : {'iso3' : 'VCT', 'site' : ''},
        'ws' : {'iso3' : 'WSM', 'site' : ''},
        'sm' : {'iso3' : 'SMR', 'site' : ''},
        'st' : {'iso3' : 'STP', 'site' : ''},
        'sa' : {'iso3' : 'SAU', 'site' : ''},
        'sn' : {'iso3' : 'SEN', 'site' : ''},
        'rs' : {'iso3' : 'SRB', 'site' : ''},
        'sc' : {'iso3' : 'SYC', 'site' : ''},
        'sl' : {'iso3' : 'SLE', 'site' : ''},
        'sg' : {'iso3' : 'SGP', 'site' : 'https://sg.morningstar.com/sg/'},
        'sx' : {'iso3' : 'SXM', 'site' : ''},
        'sk' : {'iso3' : 'SVK', 'site' : ''},
        'si' : {'iso3' : 'SVN', 'site' : ''},
        'sb' : {'iso3' : 'SLB', 'site' : ''},
        'so' : {'iso3' : 'SOM', 'site' : ''},
        'za' : {'iso3' : 'ZAF', 'site' : ''},
        'gs' : {'iso3' : 'SGS', 'site' : ''},
        'ss' : {'iso3' : 'SSD', 'site' : ''},
        'es' : {'iso3' : 'ESP', 'site' : 'https://www.morningstar.es/es/'},
        'lk' : {'iso3' : 'LKA', 'site' : ''},
        'sd' : {'iso3' : 'SDN', 'site' : ''},
        'sr' : {'iso3' : 'SUR', 'site' : ''},
        'sj' : {'iso3' : 'SJM', 'site' : ''},
        'se' : {'iso3' : 'SWE', 'site' : 'https://www.morningstar.se/se/'},
        'ch' : {'iso3' : 'CHE', 'site' : 'https://www.morningstar.ch/ch/'},
        'sy' : {'iso3' : 'SYR', 'site' : ''},
        'tw' : {'iso3' : 'TWN', 'site' : ''},
        'tj' : {'iso3' : 'TJK', 'site' : ''},
        'tz' : {'iso3' : 'TZA', 'site' : ''},
        'th' : {'iso3' : 'THA', 'site' : ''},
        'tl' : {'iso3' : 'TLS', 'site' : ''},
        'tg' : {'iso3' : 'TGO', 'site' : ''},
        'tk' : {'iso3' : 'TKL', 'site' : ''},
        'to' : {'iso3' : 'TON', 'site' : ''},
        'tt' : {'iso3' : 'TTO', 'site' : ''},
        'tn' : {'iso3' : 'TUN', 'site' : ''},
        'tr' : {'iso3' : 'TUR', 'site' : ''},
        'tm' : {'iso3' : 'TKM', 'site' : ''},
        'tc' : {'iso3' : 'TCA', 'site' : ''},
        'tv' : {'iso3' : 'TUV', 'site' : ''},
        'ug' : {'iso3' : 'UGA', 'site' : ''},
        'ua' : {'iso3' : 'UKR', 'site' : ''},
        'ae' : {'iso3' : 'ARE', 'site' : ''},
        'gb' : {'iso3' : 'GBR', 'site' : ''},
        'um' : {'iso3' : 'UMI', 'site' : ''},
        'us' : {'iso3' : 'USA', 'site' : 'https://www.morningstar.com/'},
        'uy' : {'iso3' : 'URY', 'site' : ''},
        'uz' : {'iso3' : 'UZB', 'site' : ''},
        'vu' : {'iso3' : 'VUT', 'site' : ''},
        've' : {'iso3' : 'VEN', 'site' : ''},
        'vn' : {'iso3' : 'VNM', 'site' : ''},
        'vg' : {'iso3' : 'VGB', 'site' : ''},
        'vi' : {'iso3' : 'VIR', 'site' : ''},
        'wf' : {'iso3' : 'WLF', 'site' : ''},
        'eh' : {'iso3' : 'ESH', 'site' : ''},
        'ye' : {'iso3' : 'YEM', 'site' : ''},
        'zm' : {'iso3' : 'ZMB', 'site' : ''},
        'zw' : {'iso3' : 'ZWE', 'site' : ''},

 }






"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import datetime
import json
import re

import pandas as pd
import requests
from bs4 import BeautifulSoup

from PKNSETools.morningstartools.error import no_site_error, not_200_response
from PKNSETools.morningstartools.search import search_funds, token_investment_strategy
from PKNSETools.morningstartools.security import Security
from PKDevTools.classes.Utils import random_user_agent


class Funds(Security):
    """
    Main class to access data about funds and etf, inherit from Security class

    Args:
        term (str): text to find a fund, can be a name, part of a name or the isin of the funds
        country (str) : text for code ISO 3166-1 alpha-2 of country, should be '' for etf
        pageSize (int): number of funds to return
        itemRange (int) : index of funds to return (must be inferior to PageSize)
        proxies = (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Funds('0P0000712R', "ca", 9, 0)
        >>> Funds('bond', "uk", 25, 2)

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no fund found

    """

    def __init__(self, term = None, country: str = "", pageSize : int =1, itemRange: int = 0, filters: dict = {}, proxies: dict = {}):
        
        super().__init__(term=term,asset_type='fund',country=country,pageSize=pageSize,itemRange=itemRange,filters=filters,proxies=proxies)
        
        

    def allocationMap(self):
        """    
        This function retrieves the asset allocation of the funds, index and category.

        Returns:
            dict with allocation map

        Examples:
            >>> Funds("myria", "fr").allocationMap()

        """
        return self.GetData("process/asset/v2")

    def allocationWeighting(self):
        """
        This function retrieves the Growth/Blend/Value and market capitalizaton allocation size of the funds.

        Returns:
            dict with allocation

        Examples:
            >>> Funds("myria", "fr").allocationWeighting()
        
        """
        return self.GetData("process/weighting")

    def analystRating(self):
        """
        This function retrieves the rating of the funds

        Returns:
            list of dict with ratings

            >>> Funds("RMAGX", "us").analystRating()

        """
        

        return self.GetData("parent/analystRating")

    def analystRatingTopFunds(self):
        """
        This function retrieves the rating Top funds

        Returns:
            dict with ratings

            >>> Funds("RMAGX", "us").analystRatingTopFunds()
            
        """
        
        return self.GetData("parent/analystRating/topfunds")


    def analystRatingTopFundsUpDown(self):
        """
        This function retrieves the rating funds Up Down

        Returns:
            dict with ratings

            >>> Funds("RMAGX", "us").analystRatingTopFundsUpDown()
            
        """

        return self.GetData("parent/analystRating/topfundsUpDown")


    def AnnualPerformance(self, cat):
        """
        This function retrieves the annual performance of the funds, index, category or the annual rank of the funds.

        Args:
            cat (str) : possible values are category, funds, index, rank    
        Returns:
            dict annual performance or rank

        Raises:
            ValueError : raised whenever parameter cat is not category, funds, index, or rank

        Examples:
            >>> Funds("myria", "fr").AnnualPerformance("category")
            >>> Funds("myria", "fr").AnnualPerformance("funds")
            >>> Funds("myria", "fr").AnnualPerformance("index")
            >>> Funds("myria", "fr").AnnualPerformance("rank")

        """
        no_site_error(self.code,self.name,self.country,self.site)

        cat_row = {'funds' : 0,'category' : 1, 'index' : 2, 'rank' : 3}
        if cat not in cat_row:
            raise ValueError(f"cat parameter must take one of the following value : { ', '.join(cat_row.keys())}")

        result = {}
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"
        
        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url,response)
        soup = BeautifulSoup(response.text, 'html.parser')
        #label are dates
        regex = re.compile('.*heading number')
        label_list = soup.find(id='returnsCalenderYearDiv').find_all('td', {"class": regex})
        #funds performance, category performance, index performance, rank in category
        regex = re.compile('.*value number')
        #values
        value_list = soup.find(id='returnsCalenderYearDiv').find_all('td', {"class": regex})

        regex = re.compile('-|\/')
        #first col is nothing
        for i in range(1, len(label_list)):
            label = label_list[i].text
            #if today

            if regex.search(label):
                label = 'current'
            #add category to label
            if label:
                
                result[f'{cat}_annual_performance_{label}'] = value_list[i+(cat_row[cat])*(len(label_list)-1)-1].text

        return result

    def benchmark(self):
        """
        This function retrieves the benchmark name of the funds.

        Returns:
            str benchmark name

        Examples:
            >>> Funds("myria", "fr").benchmark()

        """
        return self.referenceIndex("benchmark")

    def carbonMetrics(self):
        """
        This function retrieves the carbon metrics of the funds.

        Returns:
            dict with carbon metrics

        Examples:
            >>> Funds("myria", "fr").carbonMetrics()

        """

        return self.GetData("esg/carbonMetrics")

    def category(self):
        """
        This function retrieves the category name of the funds.

        Returns:
            str category name

        Examples:
            >>> Funds("myria", "fr").category()

        """
        return self.referenceIndex("category")

    def categoryAnnualPerformance(self):
        """
        This function retrieves the annual performance of the category.
  
        Returns:
            dict annual performance of the category


        Examples:
            >>> Funds("myria", "fr").categoryAnnualPerformance()

        """
        return self.AnnualPerformance('category')

    def categoryCumulativePerformance(self):
        """
        This function retrieves the cumulative performance of the category.
  
        Returns:
            dict cumulative performance of the category

        Examples:
            >>> Funds("myria", "fr").categoryCumulativePerformance()

        """
        
        return self.CumulativePerformance('category')

    def contact(self):
        """
        This function retrieves information about the asset manager.
  
        Returns:
            dict contact

        Examples:
            >>> Funds("myria", "fr").contact()

        """
        no_site_error(self.code,self.name,self.country,self.site)
        result = {}
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #page 1 - performance
        #page 4 - info about found
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=4"
        response = requests.get(url, headers=headers,proxies=self.proxies)
        not_200_response(url,response)
        soup = BeautifulSoup(response.text, 'html.parser')
        #label
        label_list = soup.find(id='managementManagementDiv').find_all('td', {"class": "col1 label"})
        #value
        value_list = soup.find(id='managementManagementDiv').find_all('td', {"class": "col2 value number"})
        for i in range(0, len(value_list)):
            label = label_list[i].text
            
            result[label] = value_list[i].text

        return result

    def costIllustration(self):
        """
        This function retrieves the cost of the funds.
  
        Returns:
            dict cost of funds

        Examples:
            >>> Funds("FOUSA00E5P", "us").costIllustration()

        """
        return self.GetData("price/costIllustration")

    def couponRange(self):
        """
        This function retrieves the coupon of the funds, index and category.
  
        Returns:
            dict coupon

        Examples:
            >>> Funds("rmagx", "us").couponRange()

        """
        return self.GetData("process/couponRange")


    def creditQuality(self):
        """
        This function retrieves the credit notation of the funds, index and category.
  
        Returns:
            dict credit notation

        Examples:
            >>> Funds("rmagx", "us").creditQuality()

        """
        return self.GetData("portfolio/creditQuality")

    def dataPoint(self, field, currency ='INR'):
        """
        This function retrieves infos about funds such as name, performance, risk metrics...

        Args:
        field (str or list) : field to find
        currency (str) : currency in 3 letters

        Returns:
            list of dict funds infos

        Example:
            >>> Funds("myria", "fr").dataPoint(['largestSector', 'Name', 'ongoingCharge'])
            >>> Funds("myria", "fr").dataPoint('SharpeM36')

        """
        return search_funds(self.code, field,self.country,10,currency, proxies=self.proxies)

    def distribution(self, period = "annual"):
        """
        This function retrieves the coupon distributed by the funds.

        Args:
            period (str) : annual or latest
        
        Raises:
            ValueError whenever the pariod parameter is not annual or latest

        Returns:
            dict distribution of coupon

        Example:
            >>> Funds("rmagx", "us").distribution("annual")
            
        """


        period_choice = ["annual", "latest"]
        if period not in period_choice:
            raise ValueError(f'period parameter can only take one of the values: {", ".join(period_choice)}')

        return self.GetData(f"distribution/{period}")



    def CumulativePerformance(self, cat):
        """
        This function retrieves the cumulative performance of funds, index and category.

        Args:
            cat (str) : possible values are category, funds, index

        Returns:
            dict cumulative performance

        Examples:
            >>> Funds("myria", "fr").CumulativePerformance("funds")
            >>> Funds("myria", "fr").CumulativePerformance("index")
            >>> Funds("myria", "fr").CumulativePerformance("category")

        """
        no_site_error(self.code,self.name,self.country,self.site)
        cat_row = {'funds' : 2,'category' : 3, 'index' : 4}
        if cat not in cat_row:
            raise ValueError(f"cat parameter must take one of the following value : { ', '.join(cat_row.keys())}")
        result = {}

        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"
        
        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url,response)
        soup = BeautifulSoup(response.text, 'html.parser')
        cumulative_performance_date = soup.find(id='returnsTrailingDiv').find('td', {"class": "titleBarNote"}).text
        result['cumulative_performance_date'] = cumulative_performance_date 
        #days
        regex = re.compile('.*label')
        label_list = soup.find(id='returnsTrailingDiv').find_all('td', {"class": regex})

        #cumulative performance cat
        regex = re.compile(f'.*col{str(cat_row[cat])} value number')
        value_list = soup.find(id='returnsTrailingDiv').find_all('td', {"class": regex})
        #loop on label
        for i in range(0, len(label_list)):
            #label
            label = label_list[i].text
            #perf funds
            result[f'{cat}_cumulative_performance_{label}'] = re.sub('[^0-9,-\.]','',value_list[i].text)
            
        return result

    def equityStyle(self):
        """
        This function retrieves the equity style of the funds and category.

        Returns:
            dict equity style

        Examples:
            >>> Funds("myria", "fr").equityStyle()

        """
        return self.GetData("process/stockStyle/v2")
    
    def equityStyleBoxHistory(self):
        """
        This function retrieves the equity style history of the funds

        Returns:
            dict equity style history

        Examples:
            >>> Funds("myria", "fr").equityStyleBoxHistory()

        """
        return self.GetData("process/equityStyleBoxHistory")

    def esgData(self):
        """
        This function retrieves ESG data of the funds and category

        Returns:
            dict ESG data

        Examples:
            >>> Funds("myria", "fr").esgData()

        """

        return self.GetData("esg/v1")

    def factorProfile(self):
        """
        This function retrieves the factor profile of the funds, index and category

        Returns:
            dict factor profile

        Examples:
            >>> Funds("myria", "fr").factorProfile()

        """
        return self.GetData("factorProfile")

    def feeLevel(self):
        """        
        This function retrieves the fees of the fund compare to its category.

        Returns:
            dict fees

        Examples:
            >>> Funds("rmagx", "us").feeLevel()

        """
        return self.GetData("price/feeLevel")
    
    def feeMifid(self,currency="EUR"):
        """        
        This function retrieves the fees of the fund.

        Returns:
            dict fees

        Examples:
            >>> Funds("myria", "fr").feeMifid()

        """
        return self.ltData("Mifid",currency=currency)


    def fees(self):
        """        
        This function retrieves the fees of the fund (by scraping pages);

        Returns:
            dict fees

        Examples:
            >>> Funds("myria", "fr").fees()

        """
        no_site_error(self.code,self.name,self.country,self.site)
        result = {}
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=5"
        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url,response)
        soup = BeautifulSoup(response.text, 'html.parser')
        if soup.find(id='managementFeesDiv') == None:
            return {}
        #label
        label_list =soup.find(id='managementFeesDiv').find_all('td', {"class": "label"})
        #value
        value_list = soup.find(id='managementFeesDiv').find_all('td', {"class": "value number"}) + soup.find(id='managementFeesDiv').find_all('td', {"class": "value number jdpa"})
        for i in range(0, len(value_list)):
            label = label_list[i].text
            result[label] = re.sub('(\\n +)|(\\n)','',value_list[i].text)

        return result

    def financialMetrics(self):
        """        
        This function retrieves the final metrics of the funds and category.

        Returns:
            dict financial metrics

        Examples:
            >>> Funds("rmagx", "us").financialMetrics()

        """
        return self.GetData("process/financialMetrics")


    def fixedIncomeStyle(self):
        """        
        This function retrieves the fixed income style of the funds and category.

        Returns:
            dict fixed income style

        Examples:
            >>> Funds("rmagx", "us").fixedIncomeStyle()

        """

        return self.GetData("process/fixedIncomeStyle")

    def fixedincomeStyleBoxHistory(self):
        """        
        This function retrieves the fixed income style history of the funds.

        Returns:
            dict fixed income style

        Examples:
            >>> Funds("rmagx", "us").fixedincomeStyleBoxHistory()

        """
        return self.GetData("process/fixedincomeStyleBoxHistory")

    def fundsAnnualPerformance(self):
        """        
        This function retrieves the annual performance of the funds.

        Returns:
            dict funds annual performance 

        Examples:
            >>> Funds("myria", "fr").fundsAnnualPerformance()

        """
        return self.AnnualPerformance('funds')

    def fundsAnnualRank(self):
        """        
        This function retrieves the annual rank of the funds in percentile.

        Returns:
            dict funds annual rank

        Examples:
            >>> Funds("myria", "fr").fundsAnnualRank()

        """
        return self.AnnualPerformance('rank')

    def fundsCumulativePerformance(self):
        """        
        This function retrieves the cumulative performance of the funds.

        Returns:
            dict funds cumulative performance

        Examples:
            >>> Funds("myria", "fr").fundsCumulativePerformance()

        """
        return self.CumulativePerformance('funds')

    def fundsQuarterlyPerformance(self):
        """        
        This function retrieves the quarterly performance of the funds.

        Returns:
            dict funds quarterly performance

        Examples:
            >>> Funds("myria", "fr").fundsCumulativePerformance()

        """
        no_site_error(self.code,self.name,self.country,self.site)
        result = {}
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #page 1 - performance
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}&tab=1"
        
        response = requests.get(url, headers=headers, proxies=self.proxies)
        not_200_response(url,response)
        soup = BeautifulSoup(response.text, 'html.parser')
        quarterly_performance_date = soup.find(id='returnsTrailingDiv').find('td', {"class": "titleBarNote"}).text
        result['quarterly_performance_date'] = quarterly_performance_date 

        #quarter label
        regex = re.compile('.*heading number')
        quarter_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #year label
        regex = re.compile('.*label')
        year_list =soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #1st Quarter 
        regex = re.compile('.*col2 value number')
        quarter_1_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #2nd Quarter
        regex = re.compile('.*col3 value number')
        quarter_2_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #3rd Quarter
        regex = re.compile('.*col4 value number')
        quarter_3_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #4th Quarter
        regex = re.compile('.*col5 value number')
        quarter_4_list = soup.find(id='returnsQuarterlyDiv').find_all('td', {"class": regex})
        #loop on year
        for i in range(0, len(year_list)):
            label = 'performance_%s_' % (year_list[i].text)
            result[label + 'quarter_1'] = quarter_1_list[i].text
            result[label + 'quarter_2'] = quarter_2_list[i].text
            result[label + 'quarter_3'] = quarter_3_list[i].text
            result[label + 'quarter_4'] = quarter_4_list[i].text
        return result

    def graphData(self):
        """        
        This function retrieves historical data of the funds.

        Returns:
            dict historical data

        Examples:
            >>> Funds("myria", "fr").graphData()
   
        """

        return self.GetData("parent/graphData")

    def historicalData(self):
        """
        This function retrieves the historical price of the funds, index and category

        Returns:
            dict with historical data

        Examples:
            >>> Funds("myria", "fr").historicalData()

        """
        return self.GetData("performance/v3", url_suffixe='')



    def historicalExpenses(self):
        """        
        This function retrieves historical expenses of the funds.

        Returns:
            dict historical expenses

        Examples:
            >>> Funds("rmagx", "us").historicalExpenses()

        """
        if self.asset_type == 'etf':
            return {}
        return self.GetData("price/historicalExpenses")

    def holdings(self, holdingType: str = 'all'):
        """        
        This function retrieves holdings of the funds.

        Args:
            holdingType (str) : paramater to select the kind of holdings; all, bond, equity or other

        Returns:
            pandas DataFrame holdings

        Raises:
            ValueError whenever the parameter is not all, bond, equity or other

        Examples:
            >>> Funds("myria", "fr").holdings("all")
            >>> Funds("myria", "fr").holdings("bond")
            >>> Funds("myria", "fr").holdings("equity")
            >>> Funds("myria", "fr").holdings("other")

        """
        holdingType_to_holdingPage = {"all" : "all", "bond" : "boldHoldingPage","equity" : "equityHoldingPage", "other": "otherHoldingPage"}
        if holdingType not in holdingType_to_holdingPage:
            raise ValueError(f'parameter holdingType must take one of the following value : {", ".join(holdingType_to_holdingPage.keys())}  ')
    
        if holdingType == "all":
            return pd.DataFrame(self.position()["equityHoldingPage"]["holdingList"] + 
            self.position()["boldHoldingPage"]["holdingList"] + self.position()["otherHoldingPage"]["holdingList"] )
        else:
           return pd.DataFrame(self.position()[holdingType_to_holdingPage[holdingType]]["holdingList"])

    def indexAnnualPerformance(self):
        """
        This function retrieves the annual performance of the index.
  
        Returns:
            dict annual performance of the index


        Examples:
            >>> Funds("myria", "fr").indexAnnualPerformance()

        """
        return self.AnnualPerformance('index')

    def indexCumulativePerformance(self):
        """
        This function retrieves the cumulative performance of the index.
  
        Returns:
            dict cumulative performance of the index


        Examples:
            >>> Funds("myria", "fr").indexCumulativePerformance()

        """
        
        return self.CumulativePerformance('index')

    def investmentStrategy(self):
        """
        This function retrieves the investment strategy.
  
        Returns:
            dict investment strategy

        Examples:
            >>> Funds("LU0823421689").investmentStrategy()

        """
        bearer_token = token_investment_strategy()
        #header with bearer token
        headers = {
                    'authorization': f'Bearer {bearer_token}',
                    }
        return self.GetData("morningstarTake/investmentStrategy",headers=headers)
        
    def investmentLookup(self,currency="EUR"):
        """        
        This function gives details about fund investment.

        Returns:
            dict fund investment

        Examples:
            >>> Funds("myria", "fr").investmentLookup()

        """
        return self.ltData("investmentTypeLookup",currency=currency)
    
    def marketCapitalization(self):
        """
        This function retrieves the marketCapitalization breakdown of the funds, category and index.
  
        Returns:
            dict market capitalization

        Examples:
            >>> Funds("myria", "fr").marketCapitalization()

        """
        return self.GetData("process/marketCap")


    def maturitySchedule(self):
        """
        This function retrieves the maturity breakdown of the funds and category.
  
        Returns:
            dict maturity

        Examples:
            >>> Funds("rmagx", "us").maturitySchedule()

        """
        return self.GetData("process/maturitySchedule")


    def maxDrawDown(self, year = 3):
        """
        This function retrieves the max drawdown of the funds, index and category.
  
        Args:
            year (int) : period of calculation in year

        Returns:
            dict max drawdown

        Raises:
            TypeError whenever the parameter year is not an integer

        Examples:
            >>> Funds("myria", "fr").maxDrawDown()

        """

        if not isinstance(year, int):
            raise TypeError('year parameter should be an integer')


        return self.GetData("performance/marketVolatilityMeasure", params = {"year": year})

    def morningstarAnalyst(self):
        """
        This function retrieves the raiting of MorningStar analyst.
  
        Returns:
            dict rating

        Examples:
            >>> Funds("rmagx", "us").morningstarAnalyst()
   
        """

        return self.GetData("morningstarAnalyst")

    def multiLevelFixedIncomeData(self, primary = "superEffectiveDuration", secondary = "superSector.weight"):
        """
        This function retrieves the exposures of fixed income

        Args:
            primary (str) : can be superEffectiveDuration, superYieldToWorst, creditQuality
            secondary (str) : can be superSector.weight, region.weight, creditQuality.weight

        Returns:
            dict exposures

        Raises:
            ValueError whenever the primary and seconday paramater is not a value expected

        Examples:
            >>> Funds("rmagx", "us").multiLevelFixedIncomeData()

        """

        primary_choice = ["superEffectiveDuration","superYieldToWorst", "creditQuality"]
        secondary_choice = ["superSector.weight", "region.weight", "creditQuality.weight"]

        if primary not in primary_choice:
            raise ValueError(f'primary parameter can only take one of the values: {", ".join(primary_choice)}')
        
        if secondary not in secondary_choice:
            raise ValueError(f'secondary parameter can only take one of the values: {", ".join(secondary_choice)}')

        if primary == "creditQuality" and secondary == "creditQuality.weight":
            raise ValueError(f'primary and secondary parameters cannot be both credit quality')

        return self.GetData("multiLevelFixedIncomeData", params = {"primary": primary,"secondary": secondary})

    def nav(self,start_date,end_date,frequency="daily"):
        """
        This function retrieves the NAV of the funds

        Args:
            start_date (datetime) : start date to get nav
            end_date (datetime) : end date to get nav
            frequency (str) : can be daily, weekly, monthly

        Returns:
            list of dict with nav

            >>> Funds("RMAGX", "us").nav(datetime.datetime.today()- datetime.timedelta(30),datetime.datetime.today())

        Raises:
            TypeError: raised whenever the parameter type is not the type expected
            ValueError : raised whenever the parameter is not valid or no funds found

        """

        return self.TimeSeries(["nav","totalReturn"],
                                start_date=start_date,end_date=end_date,frequency=frequency)
            
        



    def objectiveInvestment(self):
        """        
        This function retrieves the objective of investment of the fund (by scraping pages);

        Returns:
            str objective investment

        Examples:
            >>> Funds("myria", "fr").objectiveInvestment()

        """
        
        no_site_error(self.code,self.name,self.country,self.site)
        #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #Page 1 - overview
        #url page overview
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}"
        #get HTML page overview
        response = requests.get(url, headers=headers, proxies=self.proxies)
        #if page not found

        not_200_response(url,response)
        #html page as soup
        soup = BeautifulSoup(response.text, 'html.parser')
        #investment objective funds
        return soup.find(id='overviewObjectiveDiv').find('td', {"class": "value text"}).text

    def otherFee(self):
        """
        This function retrieves the other fee of the etf

        Returns:
            dict fees

        Examples:
            >>> Funds("American Century Foc Dynmc Gr ETF").otherFee()
            
            {'expenseWaiver': False, 'expenseReimbursement': None, 'expirationDate': None, 'expenseWaivers': None}
                    
        """
        return self.GetData("price/otherFee")


    def ownershipZone(self):
        """        
        This function retrieves ownershipZone of the funds, index and category.

        Returns:
            dict ownershipZone

        Examples:
            >>> Funds("myria", "fr").ownershipZone()

        """

        return self.GetData("process/ownershipZone")

    def parentMstarRating(self):

        """
        This function retrieves the raiting of parent by MorningStar analyst.
  
        Returns:
            list of dict rating

        Examples:
            >>> Funds("rmagx", "us").parentMstarRating()

        """

        return self.GetData("parent/parentMstarRating")

    def parentSummary(self):
        """
        This function retrieves info about the parent.
  
        Returns:
            dict parent info

        Examples:
            >>> Funds("rmagx", "us").parentSummary()

        """
        return self.GetData("parent/parentSummary")

    def people(self):
        """
        This function retrieves info about people who works in the company.
  
        Returns:
            dict people info

        Examples:
            >>> Funds("rmagx", "us").people()

        """
        return self.GetData("people")

    def position(self):
        """
        This function retrieves the hodings of the funds.
  
        Returns:
            dict holdings

        Examples:
            >>> Funds("myria", "fr").position()
  
        """

        return self.GetData("portfolio/holding/v2", params = {"premiumNum" : 10000, "freeNum" : 10000})

    def proxyVotingManagement(self):
        """
        This function retrieves the vote of management.
  
        Returns:
            dict vote

        Examples:
            >>> Funds("rmagx", "us").proxyVotingManagement()

        """
        return self.GetData("people/proxyVoting/management")
    

    def proxyVotingShareHolder(self):
        """
        This function retrieves the vote of shareholders.
  
        Returns:
            dict vote

        Examples:
            >>> Funds("rmagx", "us").proxyVotingShareHolder()

        """
        return self.GetData("people/proxyVoting/shareHolder")

    def productInvolvement(self):
        """
        This function retrieves the involvement of the funds
  
        Returns:
            dict involvement

        Examples:
            >>> Funds("myria", "fr").proxyVotingShareHolder()

        """

        return self.GetData("esg/productInvolvement")


    def referenceIndex(self, index):
        """
        This function retrieves the name of the category or the benchmark

        Args:
            index (str) : possible values are benchmark, category     

        Returns:
            str category or benchmark

        Raises:
            ValueErrror whenever the index parameter is not category of benchmark

        Examples:
            >>> Funds("myria", "fr").referenceIndex("category")
            >>> Funds("myria", "fr").referenceIndex("benchmark")

        """
        no_site_error(self.code,self.name,self.country,self.site)

        index_row = {'benchmark' : 0,'category' : 1}
        if index not in index_row:
            raise ValueError(f"index parameter must take one of the following value : { ', '.join(index_row.keys())}")

                #headers random agent
        headers = {'user-agent' : random_user_agent()}
        #Page 1 - overview
        #url page overview
        url = f"{self.site}funds/snapshot/snapshot.aspx?id={self.code}"
        #get HTML page overview
        response = requests.get(url, headers=headers, proxies=self.proxies)
        #if page not found
        
        not_200_response(url,response)

        #html page as soup
        soup = BeautifulSoup(response.text, 'html.parser')
        benchmark_soup = soup.find(id='overviewBenchmarkDiv2Cols').find_all('td', {"class": "value text"})
        return benchmark_soup[index_row[index]].text

    def regionalSector(self):
        """
        This function retrieves the breakdown of the funds, category and index by region
  
        Returns:
            dict regional breakdown

        Examples:
            >>> Funds("myria", "fr").regionalSector()

        """
        return self.GetData("portfolio/regionalSector")

    def regionalSectorIncludeCountries(self):
        """
        This function retrieves the breakdown of the funds, category and index by region and country
  
        Returns:
            dict regional breakdown

        Examples:
            >>> Funds("myria", "fr").regionalSectorIncludeCountries()

        """
        return self.GetData("portfolio/regionalSectorIncludeCountries")



    def riskReturnScatterplot(self):
        """
        This function retrieves the return and standard deviation of the funds and category
  
        Returns:
            dict risk return

        Examples:
            >>> Funds("rmagx", "us").riskReturnScatterplot()

        """
        return self.GetData("performance/riskReturnScatterplot")

    def riskReturnSummary(self):
        """
        This function retrieves the return and risk summary of the funds compare to the category
  
        Returns:
            dict risk return

        Examples:
            >>> Funds("rmagx", "us").riskReturnSummary()

        """

        return self.GetData("performance/riskReturnSummary")

    def riskVolatility(self):
        """
        This function retrieves the alpha, beta, R², volatility and Sharpe ratio of the funds, category and index.

        Returns:
            dict econometrics

        Examples:
            >>> Funds("rmagx", "us").riskVolatility()

        """
        return self.GetData("performance/riskVolatility")

    def salesFees(self):
        """
        This function retrieves the sales fees of the funds

        Returns:
            dict fees

        Examples:
            >>> Funds("myria", "fr").salesFees()
        
        """
        if self.asset_type == 'etf':
            return {}
        return self.GetData("price/salesFees")

    def sector(self):
        """
        This function retrieves the sector breakdown of the funds, category and index
  
        Returns:
            dict sector breakdown

        Examples:
            >>> Funds("myria", "fr").sector()

        """
        return self.GetData("portfolio/v2/sector")
    
    def snapshot(self,currency="EUR"):
        """        
        This function returns a snapshot of the fund and asset manager.

        Returns:
            dict snapshot fund and asset manager

        Examples:
            >>> Funds("myria", "fr").snapshot()

        """
        return self.ltData("MFsnapshot",currency=currency)

    def starRatingFundAsc(self):
        """
        This function retrieves the MorningStar rating of the funds of the company by ascending order
  
        Returns:
            dict rating

        Examples:
            >>> Funds("myria", "fr").starRatingFundAsc()
 
        """
        
        return self.GetData("parent/mstarRating/StarRatingFundAsc")

    def starRatingFundDesc(self):
        """
        This function retrieves the MorningStar rating of the funds of the company by descending order
  
        Returns:
            dict rating

        Examples:
            >>> Funds("myria", "fr").starRatingFundDesc()

        """
        
        return self.GetData("parent/mstarRating/StarRatingFundDesc")

    def sustainability(self,currency="EUR"):
        """        
        This function retrieves the sustainability data of the fund.

        Returns:
            dict sustanability data

        Examples:
            >>> Funds("myria", "fr").sustainability()

        """
        return self.ltData("sustainability",currency=currency)
    
    def taxes(self):
        """
        This function retrieves the other fee of the etf

        Returns:
            dict taxes

        Examples:
            >>> Funds("American Century Foc Dynmc Gr ETF").taxes()

        """
        return self.GetData("price/taxes")

    def trailingReturn(self, duration ='daily'):
        """
        This function retrieves the trailing return of the funds of the company.

        Args:
        duration (str) : frequency of return can be daily, monthly or quarterly
        
        Returns:
            dict trailing return

        Raises:
            ValueError whenever the parameter duration is not daily, monthly or quarterly

        Example:
            >>> Funds("myria", "fr").trailingReturn("daily")

        """

        duration_choice = ["daily", "monthly", "quarterly"]
        if duration not in duration_choice:
            raise ValueError(f'duration parameter can only take one of the values: {", ".join(duration_choice)}')


        return self.GetData("trailingReturn/v2",{"duration" : duration})